package eu.dnetlib.uoamonitorservice.controllers;

import eu.dnetlib.uoaadmintoolslibrary.entities.Portal;
import eu.dnetlib.uoaadmintoolslibrary.handlers.utils.RolesUtils;
import eu.dnetlib.uoaadmintoolslibrary.services.PortalService;
import eu.dnetlib.uoamonitorservice.dao.*;
import eu.dnetlib.uoamonitorservice.entities.*;
import eu.dnetlib.uoamonitorservice.handlers.EntityNotFoundException;
import eu.dnetlib.uoaadmintoolslibrary.handlers.ForbiddenException;
import eu.dnetlib.uoamonitorservice.service.StakeholderService;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;

import org.springframework.security.access.AccessDeniedException;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.*;

import java.util.*;

@RestController
@CrossOrigin(origins = "*")
public class StakeholderController {
    private final Logger log = LogManager.getLogger(this.getClass());

    @Autowired
    private RolesUtils rolesUtils;

    @Autowired
    private StakeholderService stakeholderService;

    @Autowired
    private StakeholderDAO stakeholderDAO;

    @Autowired
    private TopicDAO topicDAO;

    @Autowired
    private CategoryDAO categoryDAO;

    @Autowired
    private SubCategoryDAO subCategoryDAO;

    @Autowired
    private SectionDAO sectionDAO;

    @Autowired
    private IndicatorDAO indicatorDAO;

    @Autowired
    private TopicController topicController;

    @Autowired
    private PortalService portalService;

    @PreAuthorize("isAuthenticated()")
    @RequestMapping(value = "/stakeholder/alias", method = RequestMethod.GET)
    public List<String> getAllReservedStakeholderAlias() {
        List<String> stakeholderAlias = this.stakeholderService.getAllAliases();
        stakeholderAlias.add("all");
        stakeholderAlias.add("default");
        stakeholderAlias.add("alias");
        return stakeholderAlias;
    }

//    @PreAuthorize("isAuthenticated()")
    @PreAuthorize("hasAnyAuthority(" +
        "@AuthorizationService.PORTAL_ADMIN, " +
        "@AuthorizationService.curator(#stakeholderFull.getType()))")
    @RequestMapping(value = "/build-stakeholder", method = RequestMethod.POST)
    public Stakeholder<Topic<Category<SubCategory<Section<Indicator>>>>> buildFullStakeholder(@RequestBody Stakeholder<Topic<Category<SubCategory<Section<Indicator>>>>> stakeholderFull) {
        log.debug("build stakeholder");
        log.debug("Alias: "+stakeholderFull.getAlias());

        Stakeholder<String> stakeholder = new Stakeholder<>(stakeholderFull);

        List<String> topics = new ArrayList<>();
        List<Topic<Category<SubCategory<Section<Indicator>>>>> topicsFull = new ArrayList<>();
        for(Topic topic : stakeholderFull.getTopics()) {
            Topic<Category<SubCategory<Section<Indicator>>>> topicFull = topicController.buildTopic(topic);
            topicsFull.add(topicFull);
            topics.add(topicFull.getId());
        }
        stakeholderFull.setTopics(topicsFull);
        stakeholder.setTopics(topics);

        Date date = new Date();
        stakeholder.setCreationDate(date);
        stakeholder.setUpdateDate(date);

        stakeholderFull.setCreationDate(date);
        stakeholderFull.setUpdateDate(date);

        Stakeholder<String> stakeholderSaved = stakeholderDAO.save(stakeholder);
        stakeholderFull.setId(stakeholderSaved.getId());

        Portal portal = portalService.getPortal(stakeholderFull.getAlias());
        if(portal == null) {
            portal = new Portal();
            portal.setPid(stakeholderFull.getAlias());
            portal.setName(stakeholderFull.getName());
            portal.setType(stakeholderFull.getType());
            portalService.insertPortal(portal);
        }

        return stakeholderFull;
        //return null;
    }

    public Stakeholder setFullEntities(Stakeholder<String> stakeholder) {
        boolean addAll = false;
        boolean addPublicAndRestricted = false;

        if(rolesUtils.hasUpdateAuthority(stakeholder.getType(), stakeholder.getAlias())) {
                addAll = true;
                addPublicAndRestricted = true;
        } else if(rolesUtils.isMember(stakeholder.getType(), stakeholder.getAlias())) {
                addPublicAndRestricted = true;
        }

        Stakeholder<Topic> stakeholderFull = new Stakeholder<>(stakeholder);

        List<Topic> topics = new ArrayList<>();

        for (String topicId: (List<String>)stakeholder.getTopics()) {
            Topic<String> topic = topicDAO.findById(topicId);
            if(topic == null) {
                // EXCEPTION - Topic not found
                throw new EntityNotFoundException("Get stakeholder: Topic with id: "+topicId+" not found (topic exists in stakeholder: "+stakeholder.getId()+")");
            }

            if((!addAll && topic.getVisibility() == Visibility.PRIVATE)
                    || (!addPublicAndRestricted && topic.getVisibility() == Visibility.RESTRICTED)) {
                continue;
            }

            Topic<Category> topicFull = new Topic<Category>(topic);

            List<Category> categories = new ArrayList<>();

            for(String categoryId : topic.getCategories()) {
                Category<String> category = categoryDAO.findById(categoryId);
                if(category == null) {
                    // EXCEPTION - Category not found
                    throw new EntityNotFoundException("Get stakeholder: Category with id: "+categoryId+" not found (category exists in topic: "+topicId+")");
                }

                if((!addAll && category.getVisibility() == Visibility.PRIVATE)
                        || (!addPublicAndRestricted && category.getVisibility() == Visibility.RESTRICTED)) {
                    continue;
                }

                Category<SubCategory> categoryFull = new Category<SubCategory>(category);

                List<SubCategory> subCategories = new ArrayList<>();

                for(String subCategoryId : category.getSubCategories()) {
                    SubCategory<String> subCategory = subCategoryDAO.findById(subCategoryId);
                    if(subCategory == null) {
                        // EXCEPTION - SubCategory not found
                        throw new EntityNotFoundException("Get stakeholder: SubCategory with id: "+subCategoryId+" not found (subCategory exists in category: "+categoryId+")");
                    }

                    if((!addAll && subCategory.getVisibility() == Visibility.PRIVATE)
                            || (!addPublicAndRestricted && subCategory.getVisibility() == Visibility.RESTRICTED)) {
                        continue;
                    }

                    SubCategory subCategoryFull = new SubCategory<Section<Indicator>>(subCategory);

                    List<Section> sectionsCharts = new ArrayList<>();

                    for(String sectionId : subCategory.getCharts()) {
                        sectionsCharts.add(getSectionFull(sectionId, subCategoryId, addAll, addPublicAndRestricted));
                    }
                    subCategoryFull.setCharts(sectionsCharts);

                    List<Section> sectionsNumbers = new ArrayList<>();

                    for(String sectionId : subCategory.getNumbers()) {
                        sectionsNumbers.add(getSectionFull(sectionId, subCategoryId, addAll, addPublicAndRestricted));
                    }
                    subCategoryFull.setNumbers(sectionsNumbers);
                    subCategories.add(subCategoryFull);
                }
                categoryFull.setSubCategories(subCategories);
                categories.add(categoryFull);
            }
            topicFull.setCategories(categories);
            topics.add(topicFull);
        }
        stakeholderFull.setTopics(topics);
        return stakeholderFull;
    }

    private Section getSectionFull(String sectionId, String subCategoryId, boolean addAll, boolean addPublicAndRestricted) {
        Section<String> section = sectionDAO.findById(sectionId);
        if (section == null) {
            // EXCEPTION - Section not found
            throw new EntityNotFoundException("Get stakeholder: Section with id: " + sectionId + " not found (section exists in subCategory: " + subCategoryId + ")");
        }

        Section sectionFull = new Section<Indicator>(section);

        List<Indicator> indicators = new ArrayList<>();
        for (String indicatorId : section.getIndicators()) {
            Indicator indicator = indicatorDAO.findById(indicatorId);
            if (indicator == null) {
                // EXCEPTION - Indicator not found
                throw new EntityNotFoundException("Get stakeholder: Indicator with id: " + indicatorId + " not found (indicator exists in section: " + sectionId + ")");
            }

            if((!addAll && indicator.getVisibility() == Visibility.PRIVATE)
                    || (!addPublicAndRestricted && indicator.getVisibility() == Visibility.RESTRICTED)) {
                continue;
            }

            indicators.add(indicator);
        }
        sectionFull.setIndicators(indicators);

        return sectionFull;
    }

    @PreAuthorize("hasAnyAuthority(" +
            "@AuthorizationService.PORTAL_ADMIN)")
    @RequestMapping(value = "/stakeholder/all", method = RequestMethod.GET)
    public List<Stakeholder> getAllStakeholders(@RequestParam(required = false) String type) {
//        log.debug("get all stakeholders" + (type != null ? " with type: "+type : ""));

        List<Stakeholder> stakeholders;
        if(type == null) {
            stakeholders = stakeholderDAO.findAll();
        } else {
            stakeholders = stakeholderDAO.findByType(type);
        }

        List<Stakeholder> stakeholdersFull = new ArrayList<>();
        for(Stakeholder stakeholder : stakeholders) {
            stakeholdersFull.add(this.setFullEntities(stakeholder));
        }

        return stakeholdersFull;
    }

    @PreAuthorize("isAuthenticated()")
    @RequestMapping(value = "/stakeholder/default", method = RequestMethod.GET)
    public List<Stakeholder> getAllDefaultStakeholders(@RequestParam(required = false) String type) {
//        log.debug("get all default stakeholders" + (type != null ? " with type: "+type : ""));

        List<Stakeholder> stakeholders;
        if(type == null) {
            stakeholders = stakeholderDAO.findByDefaultId(null);
        } else {
            stakeholders = stakeholderDAO.findByDefaultIdAndType(null, type);
        }

        List<Stakeholder> stakeholdersFull = new ArrayList<>();

        // Remove stakeholders for which i do not have authority
        if(stakeholders != null && stakeholders.size() > 0) {
            if (rolesUtils.isPortalAdmin()) {
                for(Stakeholder stakeholder : stakeholders) {
                    stakeholdersFull.add(this.setFullEntities(stakeholder));
                }
                return stakeholdersFull;
            }

            Iterator<Stakeholder> stakeholderIterator = stakeholders.iterator();
            while(stakeholderIterator.hasNext()) {
                Stakeholder stakeholder = stakeholderIterator.next();
                if(rolesUtils.isCurator(stakeholder.getType())) {
                    stakeholdersFull.add(this.setFullEntities(stakeholder));
                    continue;
                }
                stakeholderIterator.remove();
            }
        }

        return stakeholdersFull;
    }

    @RequestMapping(value = "/stakeholder", method = RequestMethod.GET)
    public List<Stakeholder> getAllRealStakeholders(@RequestParam(required = false) String type,
                                                    @RequestParam(required = false) String defaultId) {
        return stakeholderService.getStakeholdersByTypeAndRole(type, defaultId, false);
    }

    @PreAuthorize("isAuthenticated()")
    @RequestMapping(value = "/my-stakeholder", method = RequestMethod.GET)
    public List<Stakeholder> getMyRealStakeholders(@RequestParam(required = false) String type) {
        return stakeholderService.getStakeholdersByTypeAndRole(type, null, true);
    }

    @RequestMapping(value = "/stakeholder/{alias:.+}", method = RequestMethod.GET)
    public Stakeholder getStakeholder(@PathVariable("alias") String alias) {

        Stakeholder<String> stakeholder = stakeholderDAO.findByAlias(alias);
        if(stakeholder == null) {
            // EXCEPTION - Stakeholder not found
            throw new EntityNotFoundException("Get stakeholder: Stakeholder with alias: "+alias+" not found");
        }

        if(stakeholder.getDefaultId() == null && !rolesUtils.isLoggedIn()) {
            // EXCEPTION - Unauthorized
            throw new AccessDeniedException("Get stakeholder: You are not authorized (not logged in) to access stakeholder with alias: "+alias);
        }
        if(stakeholder.getDefaultId() == null && !rolesUtils.hasCreateAndDeleteAuthority(stakeholder.getType())) {
            // EXCEPTION - Access denied
            throw new ForbiddenException("Get stakeholder: You are not authorized to access stakeholder with alias: "+alias);
        }

        if((stakeholder.getVisibility() == Visibility.PRIVATE && !rolesUtils.hasUpdateAuthority(stakeholder.getType(), stakeholder.getAlias())
                || (stakeholder.getVisibility() == Visibility.RESTRICTED && !rolesUtils.hasUpdateAuthority(stakeholder.getType(), stakeholder.getAlias()) && !rolesUtils.isMember(stakeholder.getType(), stakeholder.getAlias())))) {
//            // EXCEPTION - Access denied
            List<String> topicsEmpty = stakeholder.getTopics();
            topicsEmpty.clear();
            stakeholder.setTopics(topicsEmpty);
            stakeholder.setVisibility(Visibility.PRIVATE);
            return stakeholder;
        }

        return this.setFullEntities(stakeholder);
    }

    @PreAuthorize("hasAnyAuthority("
        + "@AuthorizationService.PORTAL_ADMIN, "
        + "@AuthorizationService.curator(#_stakeholder.getType()), "
        + "@AuthorizationService.manager(#_stakeholder.getType(), #_stakeholder.getAlias()) "
    + ")")
    @RequestMapping(value = "/save", method = RequestMethod.POST)
    public Stakeholder saveStakeholder(@RequestBody Stakeholder _stakeholder) {
        log.debug("save stakeholder");
        log.debug("Alias: "+_stakeholder.getAlias() + " - Id: "+_stakeholder.getId());
        Stakeholder<String> stakeholder = new Stakeholder<>(_stakeholder);

        Date date = new Date();
        stakeholder.setUpdateDate(date);
        List<String> topics = new ArrayList<>();
        if(_stakeholder.getId() == null) {
            stakeholder.setCreationDate(date);
        } else {
            Stakeholder<String> oldStakeholder = stakeholderDAO.findById(_stakeholder.getId());
            if(oldStakeholder == null) {
                // EXCEPTION - Stakeholder not found
                throw new EntityNotFoundException("save stakeholder: Stakeholder with id: "+_stakeholder.getId()+" not found");
            }
            for(String topicId : oldStakeholder.getTopics()) {
                Topic topic = topicDAO.findById(topicId);
                if (topic == null) {
                    // EXCEPTION - Topic not found
                    throw new EntityNotFoundException("Save stakeholder: Topic with id: "+topicId+" not found (topic exists in stakeholder: "+stakeholder.getId()+")");
                }
                topics.add(topic.getId());
            }
        }
        stakeholder.setTopics(topics);
        Stakeholder<String> stakeholderSaved = stakeholderDAO.save(stakeholder);
        _stakeholder.setId(stakeholderSaved.getId());
        _stakeholder.setCreationDate(stakeholderSaved.getCreationDate());
        _stakeholder.setUpdateDate(stakeholderSaved.getUpdateDate());
        return _stakeholder;
    }

    @PreAuthorize("isAuthenticated()")
    @RequestMapping(value = "/{stakeholderId}/delete", method = RequestMethod.DELETE)
    public boolean deleteStakeholder(@PathVariable("stakeholderId") String stakeholderId) {
        log.debug("delete stakeholder");
        log.debug("Id: "+stakeholderId);

        Stakeholder<String> stakeholder = stakeholderDAO.findById(stakeholderId);
        String pid = null;

        if(stakeholder != null) {
            pid = stakeholder.getAlias();
            if(!rolesUtils.hasCreateAndDeleteAuthority(stakeholder.getType())) {
                // EXCEPTION - Access denied
                throw new ForbiddenException("Delete stakeholder: You are not authorized to delete stakeholder with id: "+stakeholderId);
            }
            topicController.deleteTree(stakeholder);
            stakeholder.setTopics(null);
            stakeholderDAO.delete(stakeholderId);
            log.debug("Stakeholder deleted!");
            Portal portal = portalService.getPortal(pid);
            if(portal != null) {
                portalService.deletePortal(portal.getId());
            }
        } else {
            // EXCEPTION - Stakeholder not found
            throw new EntityNotFoundException("Delete stakeholder: Stakeholder with id: "+stakeholderId+" not found");
        }
        return true;
    }

    @PreAuthorize("isAuthenticated()")
    @RequestMapping(value = "/{stakeholderId}/change-visibility", method = RequestMethod.POST)
    public Stakeholder changeStakeholderVisibility(@PathVariable("stakeholderId") String stakeholderId,
                                              @RequestParam("visibility") Visibility visibility, @RequestParam(required = false) Boolean propagate) {
        log.debug("change stakeholder visibility: "+visibility + " - toggle propagate: "+((propagate != null && propagate) ? "true" : "false"));
        log.debug("Stakeholder: "+stakeholderId);

        Stakeholder stakeholder = stakeholderDAO.findById(stakeholderId);
        if (stakeholder == null) {
            // EXCEPTION - Stakeholder not found
            throw new EntityNotFoundException("Change stakeholder visibility: Stakeholder with id: "+stakeholderId+" not found");
        }
        if(!rolesUtils.hasUpdateAuthority(stakeholder.getType(), stakeholder.getAlias())) {
            // EXCEPTION - Access denied
            throw new ForbiddenException("Change stakeholder visibility: You are not authorized to update stakeholder with id: "+stakeholderId);
        }

        return changeStakeholderVisibilityTree(stakeholder, visibility, propagate);
    }

    private Stakeholder changeStakeholderVisibilityTree(Stakeholder<String> stakeholder, Visibility visibility, Boolean propagate) {
        Stakeholder<Topic> stakeholderFull = new Stakeholder<>(stakeholder);
        List<Topic> topicsFull = new ArrayList<>();

        if (propagate != null && propagate) {
            for (String topicId : stakeholder.getTopics()) {
                topicsFull.add(topicController.changeVisibilityTree(topicId, visibility, propagate));
            }
        }

        stakeholder.setVisibility(visibility);
        stakeholderDAO.save(stakeholder);
        log.debug("Stakeholder toggled!");

        stakeholderFull.setVisibility(visibility);
        stakeholderFull.setTopics(topicsFull);

        return stakeholder;
    }
}
